/* Fiprad 0.6C by Peter Hellman and Roger Abrahamsson 
    Copyright (C) 1999  Peter Hellman UMPLUG

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
    Author contact: peter@umplug.org 


*/

#include"fiprad.h"

void fiprad_init(char *whoami)
{       /* This function allocates memory for global pointers, connects to SQL server etc. */
	openlog(whoami,LOG_PID, LOG_DAEMON);
	syslog(LOG_NOTICE, "daemon starting");
	if(!opcodes->nosql)
		sqlhand(SQLCONNECT); /* connect to SQL server */
	rec = (fipra_rec *) malloc(sizeof(fipra_rec)); /* Allocate memory for global pointers */
	sqlrec = (fipra_rec *) malloc(sizeof(fipra_rec));
	mask = (fipra_rec *) malloc(sizeof(fipra_rec));
	ktable = (fipra_table *) malloc(sizeof(fipra_table));
	sqltable = (struct iparr *) malloc(sizeof(struct iparr));
	sysname = (struct utsname *) malloc(sizeof(struct utsname)); 
	fd=open("/dev/fipra", O_RDWR);
	if(fd==-1){
		syslog(LOG_ERR," daemon init fail:%m");
		fiprad_exit(EXIT_ERROR);
	}
	if(uname(sysname)==-1){
		syslog(LOG_ERR,"unable to get uname");
		fiprad_exit(EXIT_ERROR);
	} 
	cmprecs();
	ioctl(fd, FIPRA_ACT, 0);
}

void fiprad_exit(int opt)
{       /* Well, and this one does the opposite */
	if(!opcodes->nosql)
		sqlhand(SQLDISCONNECT);
	free(rec); /* Free global pointers */
	free(sqlrec);
	free(mask);
	free(ktable);
	free(opcodes);
	free(sysname); 
	close(fd);
	switch(opt){
	case EXIT_NORMAL:
		syslog(LOG_NOTICE, "daemon exit normal");
		break;
	case EXIT_ERROR:
		syslog(LOG_ERR, "daemon exit on error");
		break;
	default:
		syslog(LOG_NOTICE, "daemon exit on unknown reason, this should never happen");
		break;
	}		
	closelog();
	exit(0);
}

void cmprecs(void)
{ /* Compares recs from SQL and kernel, adds and removes */
	int i, x;
	fipra_rec *ktmp;
	fipra_rec *stmp;
	ktmp=(fipra_rec *) malloc(sizeof(fipra_rec));
	stmp=(fipra_rec *) malloc(sizeof(fipra_rec));
	memset(ktmp, 0, sizeof(fipra_rec));
	memset(stmp, 0, sizeof(fipra_rec));
	getsqlrec();
	getkrec();
	memcpy(ktmp, rec, sizeof(fipra_rec));
	memcpy(stmp, sqlrec, sizeof(fipra_rec));
	for(x=0;x<=NOREC;x++){
		for(i=0;i<=NOREC;i++){
			if(stmp->rec[i]==ktmp->rec[x]){
				ktmp->rec[x]=0; 
				stmp->rec[i]=0;
			}
		}
			
	}
	
	for(i=0;i<=NOREC;i++){
		if(ktmp->rec[i]!=0)
			ioctl(fd, FIPRA_RMCB, ktmp->rec[i]);
		
	}
	for(i=0;i<=NOREC;i++){
		if(stmp->rec[i]!=0)
			ioctl(fd, FIPRA_ADDCB, stmp->rec[i]);
		
	}		
	free(stmp);
	free(ktmp);
	getsqlrec(); /* re read rec's This is a bugfix */
	getkrec();
}
	
void sleep_infinite(void)
{
	alarm(0);
	pause();
}

void inttostr(unsigned long int ip, char *buf)
{        /* Well, this bad boy converts a unsigned long to ip number in string format, in dotted q form */
	char cb1[5], cb2[5], cb3[5], cb4[5];
	int ib1, ib2, ib3, ib4;
	memset(buf, '\0', sizeof(buf));
	ib1= ((ip & 0xFF000000) >> 24);
	ib2= ((ip & 0x00FF0000) >> 16);
	ib3= ((ip & 0x0000FF00) >> 8);
	ib4= (ip & 0x000000FF);
	sprintf(cb1, "%d", ib1);
	sprintf(cb2, "%d", ib2);
	sprintf(cb3, "%d", ib3);
	sprintf(cb4, "%d", ib4);
	strcat(buf, cb1);
	strcat(buf, "."); 
	strcat(buf, cb2);
	strcat(buf, "."); 
	strcat(buf, cb3);
	strcat(buf, "."); 
	strcat(buf, cb4);
}

char *stripnl(char *str)
{       /* A standard function i incoroprated, strips \n from strings, just in case */
        char tmpstr[CHARBUF], *ptmpstr;
        int i,x;
        ptmpstr=tmpstr;
        i=strlen(str);
        if(str==NULL)
                return(NULL);
        strcpy(ptmpstr, str);
        for(x=1;x<=i;x++) {
                if(*ptmpstr=='\n'){
                        *ptmpstr='\0';
                        break;
                }
                ptmpstr++;
        }
        strcpy(str, tmpstr);
	return 0;
}

unsigned long int strtoint(char *instr)
{        /* well, this one takes any dotted q ipnumber string and returns a unsigned long */
	char buf[30];
	char byte1[3];
	char byte2[3];
	char byte3[3];
	char byte4[3];
	int i=0, b1=0,b2=0,b3=0,b4=0, temp;
        unsigned long int ip;
	strcpy(buf, instr);
	memset(byte1, '\0', sizeof(byte1));
	memset(byte2, '\0', sizeof(byte1));
	memset(byte3, '\0', sizeof(byte1));
	memset(byte4, '\0', sizeof(byte1));
	temp=0;
	while(buf[i]!='\0' && buf[i]!='.') {
		byte1[temp]=buf[i];
		i++; temp++;
	}
	i++; temp=0;
	while(buf[i]!='\0' && buf[i]!='.') {
		byte2[temp]=buf[i];
		i++; temp++;
	}
	i++; temp=0;
	while(buf[i]!='\0' && buf[i]!='.') {
		byte3[temp]=buf[i];
		i++; temp++;
	}
	i++; temp=0;
	while(buf[i]!='\0' && buf[i]!='.') {
		byte4[temp]=buf[i];
		i++; temp++;
	}
	b1=atoi(byte1);
	b2=atoi(byte2);
	b3=atoi(byte3);
	b4=atoi(byte4);
	ip=b1 << 24;
	ip=ip | (b2 << 16);
	ip=ip | (b3 << 8);
	ip=ip | b4;
	return ip;
}

void start_cycle(void)
{       
/* This is the start of a working cycle, when the daemon wakes from sleep mode, 
this is where it starts, and stops, although I just changed that so it will return to main.. */
	int i,x;
	if(opcodes->sqlfuckup) { /* do we have a previous bobo ?, if so try to re connect */
		sqlhand(SQLDISCONNECT);
		sqlhand(SQLCONNECT);
		
	}
	if(!opcodes->nosql && !opcodes->sqlfuckup) { /* All good, ping just in case */
	/* Is the SQL server still alive ? */	
		if(mysql_ping(sqlp)) {
			syslog(LOG_ERR, "lost SQL connection, dumping to file");
			opcodes->sqlfuckup=1; /* operation failed, go fuckup state */
		}
	}
	if(!opcodes->nosql && !opcodes->sqlfuckup && opcodes->hasdump)
		spool();
	if(!opcodes->nosql && !opcodes->sqlfuckup)
		cmprecs(); /* Thinking about not running cmprecs so often */
	for(i=0;i<=NOREC;i++){ /* How many c-blocks in record ? */
		if(sqlrec->rec[i]==0)
			break;
		
	}
	if(opcodes->nosql || opcodes->sqlfuckup) {
		for(x=0;x<i;x++){
			getsqltable(x);
			writefile();
		}
	} else {
		for(x=0;x<i;x++){
			getsqltable(x);
			writesqltable();
		}
	}	
}

void getkrec(void)
{ /* This one reads the fipra_rec from kernel */
	memset(rec, 0, sizeof(fipra_rec));
	if(ioctl(fd, FIPRA_BLOCKS, rec)) {
		syslog(LOG_ERR, "ioctl error");
		fiprad_exit(EXIT_ERROR);
	} 
	
}

void getsqltable(int recno)
{     /* This one enters data into the SQL fipra_table, and prepares it to be written */
	int i, x, krecno=0;
	unsigned long ipnumb=0;
	char textip[16];
	memset(ktable, 0, sizeof(fipra_table));	
	memset(sqltable, 0, sizeof(struct iparr));
		/* now that we have masks, sqlrec[i] and rec[i] may not be the same anymore*/
	if((sqlrec->rec[recno]&0xFFFFFF00)!=(rec->rec[recno]&0xFFFFFF00)){
		/* oops, we have a mismatch, loop to find correct recno */
		for(i=0;i<=NOREC;i++){
			if(sqlrec->rec[recno]==rec->rec[i])
				break;
		}
		krecno=i; /* the new correct value for kernel record */
	} else
		krecno=recno; /* if no mismatch was found */
	ioctl(fd, FIPRA_SET_READCB, rec->rec[krecno]);
	if((ioctl(fd, FIPRA_ZREADCB, ktable))!=0){
		syslog(LOG_ERR, "ioctl error, cannot read from kernelspace");
		fiprad_exit(EXIT_ERROR);
	}

	for(i=1;i<=255;i++){
		/* Calculate ip number */
		ipnumb=(rec->rec[krecno] & 0xFFFFFF00);
		ipnumb= (ipnumb | i);
		
			inttostr(ipnumb, textip);
			strcpy(sqltable->ip[i].ip, textip);
			sqltable->ipdata[i].in_bytes=ktable->ipdata[i].in_bytes; 
			sqltable->ipdata[i].out_bytes=ktable->ipdata[i].out_bytes;
	}
	if(~mask->rec[recno]<255) { /* if not full cblock */

		if((sqlrec->rec[recno]&0x000000FF)>0) { /* clear lower part */
			i=(sqlrec->rec[recno]&0x000000FF);
			for(x=0;x<i;x++) {
				sqltable->ipdata[x].in_bytes=0;
				sqltable->ipdata[x].out_bytes=0;
			}
		}
		/* clear upper part */
		i=sqlrec->rec[recno]&0x000000FF;
		for(x=(i+~mask->rec[recno]);x<256;x++){
				sqltable->ipdata[x].in_bytes=0;
				sqltable->ipdata[x].out_bytes=0;				
		}	

	}
	strcpy(sqltable->ip[0].ip, "0.0.0.0");
}

void writesqltable(void)
{ /* This is where is all gets writen to SQL server */
	int i;
	char inbytes[20];
	char outbytes[20];
	char sqlstatement[125];
	for(i=0;i<=255;i++){
		if(sqltable->ipdata[i].in_bytes || sqltable->ipdata[i].out_bytes || opcodes->lognull) {
			memset(sqlstatement,'\0', strlen(sqlstatement));
			sprintf(inbytes, "%lu", sqltable->ipdata[i].in_bytes);
			sprintf(outbytes, "%lu", sqltable->ipdata[i].out_bytes);
			strcat(sqlstatement, "INSERT INTO fipra_account(time_stamp,ip,inbytes,outbytes) 
VALUES(CURRENT_TIMESTAMP(),\'");
			strcat(sqlstatement, sqltable->ip[i].ip);
			strcat(sqlstatement, "\',");
			strcat(sqlstatement, inbytes);
			strcat(sqlstatement, ",");
			strcat(sqlstatement, outbytes);
			strcat(sqlstatement, ")");
			if(mysql_query(sqlp,sqlstatement) ){
				syslog(LOG_ERR, mysql_error(sqlp));
			        opcodes->sqlfuckup=1;
				writefile();
				return;
			}
		}
	}
}

void getsqlrec(void)
{ /* This is the one that reads wich c-blocks that shall be logged */
	int i=0;
	char sqlstatement[125];
	MYSQL_RES *query_result;
	MYSQL_ROW row; 
	memset(sqlrec, 0, sizeof(fipra_rec)); /* clear record each time */
	memset(sqlstatement, 0, sizeof(sqlstatement));
	strcat(sqlstatement, "SELECT ipblock FROM fipra_blocks WHERE router='");
	strcat(sqlstatement, sysname->nodename);
	strcat(sqlstatement, "'");
	i=mysql_query(sqlp, sqlstatement); /* Do the query */
	if(i!=0) {
		syslog(LOG_ERR, mysql_error(sqlp));
		fiprad_exit(EXIT_ERROR);
	}
	query_result=mysql_store_result(sqlp); /* Extract each row and load in fipra_rec */
	i=0;
	while((row=mysql_fetch_row(query_result))!=NULL) {
		if(i>NOREC){
			syslog(LOG_ERR, "logging too many c-blocks, change #define and recompile kernel and daemon");
			fiprad_exit(EXIT_ERROR);
		}
		
		sqlrec->rec[i]=strtoint(row[0]);
		i++;
	}
	mysql_free_result(query_result);
	if(i==0) {
		syslog(LOG_ERR, "no c-blocks to log");
		fiprad_exit(EXIT_ERROR);
	}
	getsmask(0);
}

void getsmask(int i)
{ /* NEED REWRITE!! shall be in getsqlrec */
	int x=32;
	char statement[100];
	MYSQL_RES *query_result;
	MYSQL_ROW row;
	int tmp;
	memset(mask, 0, sizeof(fipra_rec));
	memset(statement, '\0', 100);
	strcat(statement, "SELECT mask FROM fipra_blocks WHERE router='");
	strcat(statement, sysname->nodename);
	strcat(statement, "'");
	i=mysql_query(sqlp, statement); /* Do the query */
	if(i!=0) {
		syslog(LOG_ERR, mysql_error(sqlp));
		fiprad_exit(EXIT_ERROR);
	}
	query_result=mysql_store_result(sqlp); /* Extract each row and load in fipra_rec */
	i=0;
	while((row=mysql_fetch_row(query_result))!=NULL) {
		if(i>NOREC){
			syslog(LOG_ERR, "logging too many c-blocks, change #define and recompile kernel and daemon");
			fiprad_exit(EXIT_ERROR);
		}
		tmp=atoi(row[0]);


		for(tmp--;tmp>=0;tmp--) {
			if(x!=0)
				mask->rec[i]|= 1<<--x;
			else if(x==0)
				mask->rec[i]|= 1;
			
		}
		i++; 
	}
	mysql_free_result(query_result);
	if(i==0) {
		syslog(LOG_ERR, "no c-blocks to log");
		fiprad_exit(EXIT_ERROR);
	}
	

}

void usage(void)
{
	printf("*****************************************************************\n");
	printf("**             fipra daemon command line arguments             **\n");
	printf("*****************************************************************\n");
	printf("**                                                             **\n");
	printf("** -d detach from controlling tty ie. deamon mode              **\n");
	printf("** -t <time> time in seconds between measurments               **\n");
	printf("** -s <server> name of sql server, may be localhost            **\n");
	printf("** none means no sql server, logging will be to files          **\n");
	printf("** !no SQL is not operational just yet!                        **\n");
	printf("** -u <user> -p <password> username and password at sqlserver  **\n");
	printf("** -b <database> name of database to open at sqlserver         **\n");
	printf("** -z logg zero values to SQL                                  **\n");
	printf("** -w write spoolfiles to SQL and exit                         **\n");
	printf("** -v print version number and exit                            **\n");
	printf("** All these opcodes (not -d) has defaults defined in fiprad.h **\n");
	printf("*****************************************************************\n");
	
}

void sqlhand(int op)
{ /* administrativia for SQL server */
	MYSQL *con;
	switch(op) {
	case SQLCONNECT:
	/* SQL server setting */
		
		sqlp=mysql_init(NULL);
		if(opcodes->sqlserver[0]==0)
			strcpy(opcodes->sqlserver ,SQLSERVER);
		if(opcodes->sqluser[0]==0)
			strcpy(opcodes->sqluser ,SQLUSER);
		if(opcodes->sqlpwd[0]==0)
			strcpy(opcodes->sqlpwd ,SQLPASSWD);
		if(opcodes->sqldb[0]==0)
			strcpy(opcodes->sqldb ,SQLDATABASE);
		con=mysql_real_connect(sqlp, opcodes->sqlserver, opcodes->sqluser, opcodes->sqlpwd, opcodes->sqldb, 0, SQLSOCK, 0);
		if(con==NULL) {
			fputs(mysql_error(sqlp), stderr);
			fputs("\n--unable to connect to SQL server--\n", stderr);
			syslog(LOG_ERR, "unable to connect to SQL server: %s", opcodes->sqlserver);
			if(!opcodes->daemon)
				fiprad_exit(EXIT_ERROR); /* If ran from command like, as testing, exit */
			else
				opcodes->sqlfuckup=1; /* daemon mode, go fuckup state */
		} else 
			opcodes->sqlfuckup=0; /* reset fuckup status, incase this is a re connect */
		
		break;
	case SQLDISCONNECT:
		mysql_close(sqlp);
		break;
	}
}	

void writefile(void) 
{
	/* Okay, uuuh, this is the routine that loggs to a file, 
	   in case of loss of conn AKA sqlfuckup. to SQL server, or if no server 
	   is being used. It creates a file in /tmp named fipa.number
	   for every cycle. Eck, by the way, this is the first added 
	   stuff since BETA testing started hehe, gotta love some new bugs
	*/
	FILE *ffp;
	time_t unixtime;
	struct tm *tmtime;
        char timestring[20];
        char filename[50];
	char tmp[33];
	char outbuf[125];
	char inbytes[20];
	char outbytes[20];
	int i;
	unixtime=time(NULL);
        tmtime=localtime(&unixtime);
        strftime(timestring, 20, "%Y-%m-%d %H:%M:%S", tmtime);
	unixtime &= 0x0000FFFF; /* heh, Unicomatic(TM) file extension generator */
	sprintf(tmp, "%lu", unixtime);
	strcpy(filename ,"/tmp/fipra.");
	strcat(filename, tmp);
	if((ffp=fopen(filename, "w")) == NULL){
		syslog(LOG_ERR, "file create error: %m");
		fiprad_exit(EXIT_ERROR);
	}
	for(i=0;i<=255;i++){ 
		if(sqltable->ipdata[i].in_bytes || sqltable->ipdata[i].out_bytes \
			|| opcodes->lognull) {
			memset(outbuf,'\0', strlen(outbuf));
			sprintf(inbytes, "%lu", sqltable->ipdata[i].in_bytes);
			sprintf(outbytes, "%lu", sqltable->ipdata[i].out_bytes);
			strcat(outbuf, " \'");
			strcat(outbuf, timestring);
			strcat(outbuf, "\',\'");
			strcat(outbuf, sqltable->ip[i].ip);
			strcat(outbuf, "\',");
			strcat(outbuf, inbytes);
			strcat(outbuf, ",");
			strcat(outbuf, outbytes);
			strcat(outbuf, "\n");
			fputs(outbuf, ffp);
		}
	}	
	
	fclose(ffp);
	opcodes->hasdump=1;
}

void spool(void)
{ /* Time to write file dumped data to SQL, this shall be re written. 
     Too deep recursion for my taste. */
	DIR *dp;
	FILE *ffp;
	struct dirent *dirp;
	char filename[50];
	char sqlstatement[125];
	char inbuf[125];
	if((dp=opendir("/tmp"))==NULL){
		syslog(LOG_ERR, "unable to open /tmp");
		fiprad_exit(EXIT_ERROR);
	}
	while((dirp=readdir(dp))!=NULL){
#ifdef DEBUG
		printf(" Found a file named : %s\n", dirp->d_name);
#endif
		if(!strncmp(dirp->d_name, "fipra.", 6)){
			strcpy(filename, "/tmp/");
			strcat(filename, dirp->d_name);
#ifdef DEBUG
			printf(" Processed filename is : %s\n", filename);
#endif
			if((ffp=fopen(filename, "r")) == NULL){
				syslog(LOG_ERR, "file open error: %m");
				fiprad_exit(EXIT_ERROR);
			}
#ifdef DEBUG
			printf(" Opened  : %s\n", filename);
#endif
			if(mysql_ping(sqlp)) {
				syslog(LOG_ERR, "lost SQL connection, in file dump, bad!");
				opcodes->sqlfuckup=1; /* operation failed, go fuckup state */
				return;
			}
			while((fgets(inbuf, 125,ffp))!=NULL) {
#ifdef DEBUG
				printf(" inbuf is %s\n", inbuf);
#endif
				stripnl(inbuf);
				strcpy(sqlstatement, "INSERT INTO fipra_account VALUES(NULL,");
				strcat(sqlstatement, inbuf);
				strcat(sqlstatement, ")");
#ifdef DEBUG
				printf(" SQL statement is : %s\n", sqlstatement);
#endif
				if(mysql_query(sqlp,sqlstatement)){ 
					syslog(LOG_ERR, mysql_error(sqlp));
					opcodes->sqlfuckup=1;
					return;
				}
#ifdef DEBUG
				printf(" Wrote to SQL : %s\n", sqlstatement);
#endif
			}
			fclose(ffp);
			remove(filename);
		}
	}
	closedir(dp);
	syslog(LOG_NOTICE, "regained SQL, spooled back data from disk");
}

void sighand(int sig)
{ /* Need to be explained ? I think not */
	switch(sig) {
	case SIGINT :
		printf(" Terminating... \n");
		fiprad_exit(EXIT_NORMAL);
		break;
	case SIGUSR1:
		syslog(LOG_NOTICE,"daemon force write to SQL");
		start_cycle();
		break;
	case SIGUSR2:
		syslog(LOG_NOTICE,"daemon put in sleep mode");
		sleep_infinite();
		break;
	case SIGTERM:
		fiprad_exit(EXIT_NORMAL);
	case SIGALRM: /* SIGALRM has a different sig set, SA_NODEFER is set */
		start_cycle();
		break;
	}
}

int main(int argc, char *argv[])
{       /* UUuuuuh what is this ? Yes, this is main */
	int i=0;
	struct sigaction sigs, cycle_sigs;
	pid_t pid;
	
	/* signal setting */
	sigs.sa_handler=sighand;
	sigemptyset(&sigs.sa_mask);
	sigs.sa_flags=0;
	cycle_sigs.sa_handler=sighand;
	sigemptyset(&cycle_sigs.sa_mask);
	cycle_sigs.sa_flags=SA_NODEFER; /* Probably not needed anymore, but, who cares */
	sigaction(SIGINT, &sigs, 0); /* Terminate clean from non deamon mode */
	sigaction(SIGUSR1, &sigs, 0);  /* Start cycle now */
	sigaction(SIGUSR2, &sigs, 0); /* Sleep infinitive */
	sigaction(SIGTERM, &sigs, 0); /* TERM clean */
	sigaction(SIGALRM, &cycle_sigs, 0); /* Start cycle now */

	/* opcodes is the structre that will contain command line options */
	opcodes = (struct ops*) malloc(sizeof(struct ops));
	memset(opcodes, 0, sizeof(struct ops)); 
	
	/* Get opcodes */
	/* d deamon, v version, s severname t sleep time */
	 while((i=getopt(argc, argv ,"zwt:dvs:u:p:b:"))!=-1) {
		switch(i){
		case 't':
			opcodes->asec=atoi(optarg);
			break;
		case 'd':
			opcodes->daemon=1;
			break;
		case 'v':
			printf(VERSION);
			exit(0);
		
		case 's':
			strcpy(opcodes->sqlserver, optarg);
			break;
		case 'u':
			strcpy(opcodes->sqluser, optarg); 	       
			break;
		case 'p':
			strcpy(opcodes->sqlpwd, optarg);
			break;
		case 'b':
			strcpy(opcodes->sqldb, optarg);
			break;
		case 'z':
			opcodes->lognull=1;
			break;
		case '?':
			usage();
			exit(0);
		case 'w':
			sqlhand(SQLCONNECT);
			spool();
			sqlhand(SQLDISCONNECT);
			exit(0);
		}
	 }
	 
	 
	 if(!strcmp("none", opcodes->sqlserver)) 
		 opcodes->nosql=1;

	 if(opcodes->asec==0)
		 opcodes->asec=DEFAULT_TIME;
	 fiprad_init(argv[0]); /* sending argv[0] to get the whoami for openlog() */
	 if(opcodes->daemon){ /* Apperenly we are going in daemon mode, thank you. */
		 if((pid=fork())<0){
			 fputs("Fork error\n", stderr);
			 fiprad_exit(EXIT_ERROR);
		 } else if(pid!=0)
			 exit(0); /* Parent goes bye-bye */
		 setsid(); /* We are now a daemon, setting session leader to prevent zombification hehe */
	 }
	 
	 start_cycle(); /* Here we go, into the fire, you go boy */
	 while(1){ /* The eternal loop of a alarm clock and sleep, kind of like my life */
		 alarm(opcodes->asec);
#ifdef DEBUG
		 printf(" WENT TO SLEEP\n");
#endif
		 pause();
	 }
	 /* This is as far as we go, the daemon should now operate on signals only */
	 syslog(LOG_ERR, "oops, programming error");
	 fiprad_exit(EXIT_ERROR);
	 return 0;	
}
















